import { LiquidityPoolJsonInfo } from '@raydium-io/raydium-sdk';
import { Connection, Transaction, SystemProgram, LAMPORTS_PER_SOL, TransactionMessage, VersionedTransaction, ComputeBudgetProgram, PublicKey } from '@solana/web3.js';
import { getAssociatedTokenAddress, createAssociatedTokenAccountInstruction, createTransferInstruction, getAccount, getMint } from "@solana/spl-token";
import RaydiumSwap from '../server/RaydiumSwap';
import { convertKeypair, convertNumber, convertPubkey, MICRO_LAMPORTS_PER_LAMPORTS, MINIMUM_TRANSACTION_FEE_IN_SOL } from '../utils';
import { getSimulationComputeUnits } from '@solana-developers/helpers';
import { BN } from '@coral-xyz/anchor';

const MAX_RETRIES = 20;
export class SolanaSwapClient {
  rpc: string = 'https://api.mainnet-beta.solana.com';
  connection: Connection;
  poolKeys: LiquidityPoolJsonInfo[];

  constructor(poolKeys: LiquidityPoolJsonInfo[], rpc: string = 'https://api.mainnet-beta.solana.com') {
    this.rpc = rpc;
    this.connection = new Connection(this.rpc);
    if (!poolKeys || poolKeys.length === 0) {
      throw new Error('Pool keys not loaded');
    }
    this.poolKeys = poolKeys;

    if (!this.poolKeys || this.poolKeys.length === 0) {
      throw new Error('Pool keys not loaded');
    }
  }

  public async swap(
    fromPrivKeyBs58: string,
    tokenIn: string,
    tokenOut: string,
    amountIn: string,
    slippage: string = "5", // 5%
    priceImpact: string | null = null,
    fee: string | null = null,
    feePercent: string,
    feeRecipient: PublicKey,
  ): Promise<{
    txid: string,
    amountOut: number,
    fee: number
  }> {
    const owner = convertKeypair(fromPrivKeyBs58);
    const amountInSols = convertNumber(amountIn);

    const raydiumSwap = new RaydiumSwap(this.rpc, owner, this.poolKeys);

    const poolInfo = raydiumSwap.findPoolInfoForTokens(tokenIn, tokenOut);
    if (!poolInfo) {
      throw new Error("Pool info not found");
    }

    const transactionInfo = await raydiumSwap.getSwapTransaction(
      tokenOut,
      amountInSols,
      poolInfo,
      slippage,
      priceImpact,
      fee,
      feePercent,
      feeRecipient,
    );

    const txid = await raydiumSwap.sendLegacyTransaction(
      transactionInfo.transaction as Transaction,
      MAX_RETRIES
    );

    const latestBlockhash = await this.connection.getLatestBlockhash();
    await this.connection.confirmTransaction({
      signature: txid,
      blockhash: latestBlockhash.blockhash,
      lastValidBlockHeight: latestBlockhash.lastValidBlockHeight,
    }, "confirmed");

    const tx = await this.connection.getTransaction(txid, { "commitment": "confirmed", "maxSupportedTransactionVersion": 1 });
    let amountOut = 0;
    if (tokenOut === "So11111111111111111111111111111111111111112") {
      amountOut = transactionInfo.amountOut;
    } else {
      const preTokenOutBalance = tx?.meta?.preTokenBalances?.find(obj => obj.mint === tokenOut && obj.owner === owner.publicKey.toBase58());
      const postTokenOutBalance = tx?.meta?.postTokenBalances?.find(obj => obj.mint === tokenOut && obj.owner === owner.publicKey.toBase58());
      const postBalance = new BN(postTokenOutBalance?.uiTokenAmount.amount ?? "0");
      const preBalance = new BN(preTokenOutBalance?.uiTokenAmount.amount ?? "0");
      const decimals = postTokenOutBalance?.uiTokenAmount.decimals ?? 0;
      amountOut = postBalance.sub(preBalance).toNumber() / 10 ** decimals;
    }

    return {
      txid,
      amountOut,
      fee: transactionInfo.fee
    };
  }

  public async calcAmountOut(
    fromPrivKeyBs58: string,
    tokenIn: string,
    tokenOut: string,
    amountIn: string,
    slippage: string = "5", // 5%
    priceImpact: string | null = null,
  ): Promise<{
    amountIn: number;
    amountOut: number;
  }> {
    const owner = convertKeypair(fromPrivKeyBs58);
    const amountInSols = convertNumber(amountIn);

    const raydiumSwap = new RaydiumSwap(this.rpc, owner, this.poolKeys);

    const poolInfo = raydiumSwap.findPoolInfoForTokens(tokenIn, tokenOut);
    if (!poolInfo) {
      throw new Error("Pool info not found");
    }

    const directionIn = poolInfo.quoteMint.toString() == tokenOut;
    const amountOut = await raydiumSwap.calcAmountOut(
      poolInfo,
      amountInSols,
      directionIn,
      slippage,
      priceImpact,
    );

    const amountOutParsed =
      amountOut.amountOut.raw.toNumber() /
      10 ** amountOut.amountOut.currency.decimals;

    return {
      amountIn: amountInSols,
      amountOut: amountOutParsed,
    };
  }
}

export async function transferSpl(
  connection: Connection,
  fromPrivKeyBs58: string,
  token: string,
  amount: string,
  receiver: string,
  fee: string | null = null,
): Promise<string> {
  const to = convertPubkey(receiver);
  const from = convertKeypair(fromPrivKeyBs58);
  let amountUi = convertNumber(amount);

  const fromAssociatedTokenAccount = await getAssociatedTokenAddress(
    convertPubkey(token),
    from.publicKey,
  );
  const toAssociatedTokenAccount = await getAssociatedTokenAddress(
    convertPubkey(token),
    to,
  );

  const tx = new Transaction();

  try {
    const toAccountInfo = await connection.getAccountInfo(toAssociatedTokenAccount);
    if (!toAccountInfo) {
      tx.add(
        createAssociatedTokenAccountInstruction(
          from.publicKey,
          toAssociatedTokenAccount,
          to,
          convertPubkey(token)
        )
      );
    }
  } catch (e) {
    tx.add(
      createAssociatedTokenAccountInstruction(
        from.publicKey,
        toAssociatedTokenAccount,
        to,
        convertPubkey(token)
      )
    );
  }

  // Fake instruction to calculate compute unit
  tx.add(
    createTransferInstruction(
      fromAssociatedTokenAccount,
      toAssociatedTokenAccount,
      from.publicKey,
      1,
    )
  );

  // Calculate the transaction fee
  let transferFee = MINIMUM_TRANSACTION_FEE_IN_SOL;

  // Get temporary units
  tx.instructions.unshift(ComputeBudgetProgram.setComputeUnitPrice({
    microLamports: MICRO_LAMPORTS_PER_LAMPORTS
  }))
  const units = await getSimulationComputeUnits(connection, tx.instructions, from.publicKey, []);
  tx.instructions.shift(); // remove fake setComputeUnitPrice
  tx.instructions.pop(); // remove fake transfer instruction
  let computeUnitPrice = 0;

  if (fee && units) {
    transferFee = convertNumber(fee);
    if (transferFee >= MINIMUM_TRANSACTION_FEE_IN_SOL) {
      const diffFee = transferFee - MINIMUM_TRANSACTION_FEE_IN_SOL;
      computeUnitPrice = Math.floor((diffFee * LAMPORTS_PER_SOL / units) * MICRO_LAMPORTS_PER_LAMPORTS);
    } else {
      throw Error("Fee must be greater than 5000 lamports (0.000005 SOL)");
    }
  }
  const balanceInSol = convertNumber(((await connection.getBalance(from.publicKey)) / LAMPORTS_PER_SOL).toString());
  if (transferFee > balanceInSol) {
    throw Error("Fee is greater than balance");
  }

  const fromTokenBalance = (await getAccount(connection, fromAssociatedTokenAccount)).amount.toString()
  const decimals = 10 ** (await getMint(connection, convertPubkey(token))).decimals;

  const balanceUi = convertNumber(fromTokenBalance) / decimals;

  if (amountUi === 0) {
    amountUi = balanceUi;
    if (!units) {
      throw Error("Cannot send max balance");
    }
  }

  tx.add(
    createTransferInstruction(
      fromAssociatedTokenAccount,
      toAssociatedTokenAccount,
      from.publicKey,
      amountUi * decimals,
    )
  );

  if (computeUnitPrice > 0 && units) {
    tx.instructions.unshift(ComputeBudgetProgram.setComputeUnitPrice({
      microLamports: computeUnitPrice
    }))

    tx.instructions.unshift(ComputeBudgetProgram.setComputeUnitLimit({
      units
    }))
  }

  tx.recentBlockhash = (await connection.getLatestBlockhash()).blockhash;
  tx.feePayer = from.publicKey;
  const recoverTx = Transaction.from(tx.serialize({ requireAllSignatures: false }));

  // transaction sign
  recoverTx.sign(from);

  const signature = await connection.sendRawTransaction(recoverTx.serialize());
  return signature;
}

export async function transferSols(
  connection: Connection,
  fromPrivKeyBs58: string,
  amount: string,
  receiver: string,
  fee: string | null = null,
): Promise<string> {
  const to = convertPubkey(receiver);
  const from = convertKeypair(fromPrivKeyBs58);
  let amountInSols = convertNumber(amount);

  // Calculate the transaction fee
  let transferFee = MINIMUM_TRANSACTION_FEE_IN_SOL;
  const units = 500; // compute unit limit to send SOL
  let computeUnitPrice = 0;
  if (fee) {
    transferFee = convertNumber(fee);
    if (transferFee >= MINIMUM_TRANSACTION_FEE_IN_SOL) {
      const diffFee = transferFee - MINIMUM_TRANSACTION_FEE_IN_SOL;
      computeUnitPrice = Math.floor((diffFee * LAMPORTS_PER_SOL / units) * MICRO_LAMPORTS_PER_LAMPORTS);
    } else {
      throw Error("Fee must be greater than 5000 lamports (0.000005 SOL)");
    }
  }

  const balanceInSol = convertNumber(((await connection.getBalance(from.publicKey)) / LAMPORTS_PER_SOL).toString());

  // if amount = 0, set the `amountInSols` as all SOL balance of `from` subtract transfer fee
  if (amountInSols === 0) {
    if (transferFee > balanceInSol) {
      throw Error("Fee is greater than balance");
    }
    amountInSols = balanceInSol - transferFee;
  } else if (amountInSols > balanceInSol) {
    throw Error("Insufficient balance");
  }

  const transferInstruction = [
    SystemProgram.transfer({
      fromPubkey: from.publicKey,
      toPubkey: to,
      lamports: amountInSols * LAMPORTS_PER_SOL
    }),
  ];

  if (computeUnitPrice > 0) {
    transferInstruction.unshift(ComputeBudgetProgram.setComputeUnitPrice({
      microLamports: computeUnitPrice
    }))

    transferInstruction.unshift(ComputeBudgetProgram.setComputeUnitLimit({
      units
    }))
  }

  let { blockhash } = await connection.getLatestBlockhash();

  const message = new TransactionMessage({
    payerKey: from.publicKey,
    recentBlockhash: blockhash,
    instructions: transferInstruction,
  }).compileToV0Message();

  const transaction = new VersionedTransaction(message);

  transaction.sign([from]);

  const signature = await connection.sendRawTransaction(transaction.serialize());
  return signature;
}


